#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 parallel1.py
#
"""
This test uses the AWS python SDK boto3 to access our local ASL Workflow Engine
https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/stepfunctions.html
https://docs.aws.amazon.com/code-samples/latest/catalog/code-catalog-python-example_code-stepfunctions.html

This example is a minimal example to illustrate use of the Parallel State
https://states-language.net/spec.html#parallel-state

It is a modification of the first example from the AWS Step Functions Developer
Guide made even simpler and more self-contained by using basic Pass States.
https://github.com/awsdocs/aws-step-functions-developer-guide/blob/master/doc_source/amazon-states-language-parallel-state.md
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import boto3, time
from botocore.exceptions import ClientError

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer

ASL = """{
  "Comment": "Parallel Example.",
  "StartAt": "LookupCustomerInfo",
  "States": {
    "LookupCustomerInfo": {
      "Type": "Parallel",
      "End": true,
      "Branches": [
        {
         "StartAt": "LookupAddress",
         "States": {
           "LookupAddress": {
             "Type": "Pass",
             "Result": "testaddress@hotmail.com",
             "End": true
           }
         }
       },
       {
         "StartAt": "LookupPhone",
         "States": {
           "LookupPhone": {
             "Type": "Pass",
             "Comment": "Fake number from https://fakenumber.org",
             "Result": "+44 1632 960837",
             "End": true
           }
         }
       }
      ]
    }
  }
}"""

items = ['{"lambda":"Success"}']

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name="parallel1")

    # Initialising OpenTracing. It's important to do this before the boto3.client
    # call as create_tracer "patches" boto3 to add the OpenTracing hooks.
    create_tracer("parallel1", {"implementation": "Jaeger"})

    # Initialise the boto3 client setting the endpoint_url to our local
    # ASL Workflow Engine
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session.client
    sfn = boto3.client("stepfunctions", endpoint_url="http://localhost:4584")
    state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:parallel1"

    def create_state_machines():
        # Create state machine using a dummy roleArn. If it already exists an
        # exception will be thrown, we ignore that but raise other exceptions.
        try:
            response = sfn.create_state_machine(
                name="parallel1", definition=ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )
        except sfn.exceptions.StateMachineAlreadyExists as e:
            #print(e.response)

            response = sfn.update_state_machine(
                stateMachineArn=state_machine_arn,
                definition=ASL
            )
        except ClientError as e:
            logger.error(e)


    create_state_machines()

    # Loop through items invoking a new state machine execution for each item
    for item in items:
        try:
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except sfn.exceptions.StateMachineDoesNotExist as e:
            logger.info(e)

            create_state_machines()
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except ClientError as e:
            logger.error(e)

    time.sleep(1)  # Give OpenTracing a chance to flush buffer before exiting




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 blocking_processor_state_machine.py
#
"""
This simple state machine is used to test "blocking" processors, e.g.
workers_blocking, workers_blocking_asyncio, etc. These deliberately simulate
processors/workers that take a long time to run and are used both to test the
approach used to launch the actual processing of those processors in a
ThreadPoolExecutor and also how the ASL Engine responds to long-lived processing.
Long running/blocking processors are know to have issues if their heartbeat
interval is less than the duration of the processing due to blocking the
event loop so launching the long running processing as a Task in a
ThreadPoolExecutor allows the event loop to continue running unblocked and
still responding to heartbeats and this is what we wish to test here.
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import boto3, time
from botocore.exceptions import ClientError

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer

ASL = """{
    "Comment": "Test Step Function",
    "StartAt": "StartState",
    "States": {
        "StartState": {
            "Type": "Pass",
            "Next": "SimpleBlockingProcessor"
        },
        "SimpleBlockingProcessor": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:SimpleBlockingProcessor",
            "Next": "EndState"
        },
        "EndState": {
            "Type": "Pass",
            "End": true
        }
    }
}"""


items = ['{"lambda":"Success"}']

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name='blocking_processor_state_machine')

    # Initialising OpenTracing. It's important to do this before the boto3.client
    # call as create_tracer "patches" boto3 to add the OpenTracing hooks.
    create_tracer("blocking_processor_state_machine", {"implementation": "Jaeger"})

    # Initialise the boto3 client setting the endpoint_url to our local
    # ASL Workflow Engine
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session.client
    sfn = boto3.client("stepfunctions", endpoint_url="http://localhost:4584")
    state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:blocking_processor_state_machine"

    def create_state_machines():
        # Create state machine using a dummy roleArn. If it already exists an
        # exception will be thrown, we ignore that but raise other exceptions.
        try:
            response = sfn.create_state_machine(
                name="blocking_processor_state_machine", definition=ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )
        except sfn.exceptions.StateMachineAlreadyExists as e:
            #print(e.response)

            response = sfn.update_state_machine(
                stateMachineArn=state_machine_arn,
                definition=ASL
            )
        except ClientError as e:
            logger.error(e)


    create_state_machines()

    # Loop through items invoking a new state machine execution for each item
    for item in items:
        try:
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except sfn.exceptions.StateMachineDoesNotExist as e:
            logger.info(e)

            create_state_machines()
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except ClientError as e:
            logger.error(e)

    time.sleep(1)  # Give OpenTracing a chance to flush buffer before exiting




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
This creates an OpenTracing tracer instance. It will default to the null
opentracing tracer and use JaegerTracing if configured to do so.
This module also "patches" the AWS boto3 Python SDK to allow tracing to be
enabled on clients in a fairly transparent non-intrusive way. To do this it
is important to call create_tracer prior to any boto3.client calls, though this
is likely to be a fairly natural pattern.
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3

import opentracing  # Provides a default null opentracing.tracer instance

from asl_workflow_engine.logger import init_logging


def instrument_StartExecution(params, **kwargs):
    """
    boto3 event handler to create an OpenTracing span when Step Function
    StartExecution is called and then to inject (serialise) the span ID onto
    an HTTP header in the StepFunctions REST API so that the trace can be
    tracked through the Step Function execution and out to the Tasks it calls.

    Some background on extending boto3 via its events subsystem:
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/events.html
    https://github.com/boto/botocore/issues/902
    https://github.com/boto/botocore/blob/develop/botocore/handlers.py#L513
    """

    """
    Start an OpenTracing trace for the boto3 StartExecution request.
    https://opentracing.io/guides/python/tracers/ standard tags are from
    https://opentracing.io/specification/conventions/
    """
    with opentracing.tracer.start_active_span(
        operation_name="StartExecution",
        child_of=opentracing.tracer.active_span,
        tags={
            "component": "boto3",
            "boto3.service_name": "stepfunctions",
            "span.kind": "producer",  # Maybe "client" as it's logically RPC
            "peer.address": params["url"]
        }
    ) as scope:
        params["headers"].update(
            inject_span("http_headers", scope.span, create_tracer.logger)
        )

def patch_boto3():
    """
    Patch boto3 to enable OpenTracing. This first intercepts boto3.client
    calls, and if the client is a stepfunctions client an event handler is
    registered on the StartExecution call that creates a span and injects
    it into the REST API HTTP headers.
    """
    try:
        import boto3

        # Save the original boto3.client function to use in the wrapper.
        boto3_client = boto3.client

        def client_wrapper(*args, **kwargs):
            client = boto3_client(*args, **kwargs)

            if args[0] == "stepfunctions":
                """
                Access boto3 event system and Register instrument_StartExecution
                function to the StartExecution event before-call hook.
                """
                client.meta.events.register(
                    "before-call.stepfunctions.StartExecution",
                    instrument_StartExecution
                )
            return client

        # Patch boto3 to use client_wrapper
        boto3.client = client_wrapper

    except Exception as e:
        create_tracer.logger.warning("Failed to add Tracer to boto3: {}".format(e))

def create_tracer(service_name, config, use_asyncio=False):
    create_tracer.logger = init_logging(service_name)
    create_tracer.logger.info("Creating OpenTracing Tracer")

    # Store default tracer in case creating concrete implementation fails.
    tracer = opentracing.tracer
    if config.get("implementation") == "Jaeger":
        try:
            # import deferred until Jaeger is selected in config.
            import jaeger_client
            import tornado.ioloop

            """
            If Implementation = Jaeger get the Jaeger config from the config
            dict if available, if not present create a sane default config.
            """
            jaeger_config = config.get("config")
            if not jaeger_config:
                jaeger_config = {
                    "sampler": {
                        "type": "const",
                        "param": 1
                    },
                    "logging": False
                }

            jaeger = jaeger_client.Config(
                service_name=config.get("service_name", service_name),
                config=jaeger_config,
            )

            """
            The init_logging(log_name="tornado") is important, though a bit
            obtuse. Without it all subsequent log messages generated will
            be duplicated. The issue is that "under the covers" Jaeger uses
            the tornado https://www.tornadoweb.org async networking library.
            Tornado's IOLoop creates a log handler if necessary when it
            is started, because if there were no handler configured
            you'd never see any of its event loop exception messages.
            The default handler is created for the root logger and ends up
            resulting in duplicate messages for other logs. By explicitly
            adding a handler for the tornado logger, as the following line
            does, the logging should be correctly handled. See:
            https://stackoverflow.com/questions/30373620/why-does-ioloop-in-tornado-seem-to-add-root-logger-handler
            """
            init_logging(log_name="tornado")

            """
            If we are using asyncio we want the tracer to use the main asyncio
            event loop rather than create a new ThreadLoop (which is the default
            behaviour unless a tornado IOLoop is passed. In recent versions of
            Tornado that delegates to the asyncio event loop so getting the
            current tornado IOLoop and passing that to initialize_tracer will
            cause the tracer to use the main event loop.
            """
            if use_asyncio:
                jaeger.initialize_tracer(io_loop=tornado.ioloop.IOLoop.current())
            else:
                jaeger.initialize_tracer()

            create_tracer.logger.info("Jaeger Tracer initialised")
        except Exception as e:
            create_tracer.logger.warning("Failed to initialise Jaeger Tracer : {}".format(e))
            opentracing.tracer = tracer

    patch_boto3()

def span_context(format, carrier, logger):
    """
    Boilerplate to extract the parent OpenTracing SpanContext from the carrier.
    Log a message and start new span if we can't extract SpanContext.
    See https://opentracing.io/docs/overview/inject-extract/
    """
    try:
        span_context = opentracing.tracer.extract(format, carrier)

    except Exception:
        logger.error("Missing trace-id property, unable to deserialise "
                     "SpanContext. Will have to create new trace")
        span_context = opentracing.tracer.start_span(operation_name="dangling_process")

    return span_context

def inject_span(format, span, logger):
    """
    Boilerplate to inject the OpenTracing SpanContext into a carrier in order
    to transport it via HTTP headers, AMQP message headers or ASL Context
    https://opentracing.io/docs/overview/inject-extract/
    """
    carrier = {}
    opentracing.tracer.inject(span, format, carrier)
    logger.debug("Tracing active span: carrier {}".format(carrier))

    return carrier




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 test_state_engine_paths.py
#

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import unittest
import json

from asl_workflow_engine.state_engine_paths import apply_jsonpath
from asl_workflow_engine.asl_exceptions import *

"""
JSON document for use by examples from https://goessner.net/articles/JsonPath/
The link above is essentially the de facto definition/specification for JSONPath
from Stefan Goessner. The tests below reflect the set of examples given at
https://goessner.net/articles/JsonPath/index.html#e3 in the document.
"""
goessner = { "store": {
    "book": [ 
      { "category": "reference",
        "author": "Nigel Rees",
        "title": "Sayings of the Century",
        "price": 8.95
      },
      { "category": "fiction",
        "author": "Evelyn Waugh",
        "title": "Sword of Honour",
        "price": 12.99
      },
      { "category": "fiction",
        "author": "Herman Melville",
        "title": "Moby Dick",
        "isbn": "0-553-21311-3",
        "price": 8.99
      },
      { "category": "fiction",
        "author": "J. R. R. Tolkien",
        "title": "The Lord of the Rings",
        "isbn": "0-395-19395-8",
        "price": 22.99
      }
    ],
    "bicycle": {
      "color": "red",
      "price": 19.95
    }
  }
}

simple_array = {
    "inputs": ["index 0", "index 1"]
}

empty_array = {
    "inputs": []
}


class TestStateEnginePaths(unittest.TestCase):

    def test_goessner01(self):
        print("The authors of all books in the store.")
        result = apply_jsonpath(goessner, "$.store.book[*].author")
        print(result)
        self.assertEqual(result, ["Nigel Rees", "Evelyn Waugh", "Herman Melville", "J. R. R. Tolkien"])

    def test_goessner02(self):
        print("All authors.")
        result = apply_jsonpath(goessner, "$..author")
        print(result)
        self.assertEqual(result, ["Nigel Rees", "Evelyn Waugh", "Herman Melville", "J. R. R. Tolkien"])

    def test_goessner03(self):
        print("All things in store, which are some books and a red bicycle.")
        result = apply_jsonpath(goessner, "$.store.*")
        print(result)
        success = False
        # The "bicycle" and "book" results could be returned in either order
        # in the result list so need to check both
        if (result == [[{"category": "reference", "price": 8.95, "author": "Nigel Rees", "title": "Sayings of the Century"}, {"category": "fiction", "price": 12.99, "author": "Evelyn Waugh", "title": "Sword of Honour"}, {"category": "fiction", "isbn": "0-553-21311-3", "price": 8.99, "author": "Herman Melville", "title": "Moby Dick"}, {"category": "fiction", "isbn": "0-395-19395-8", "price": 22.99, "author": "J. R. R. Tolkien", "title": "The Lord of the Rings"}], {"price": 19.95, "color": "red"}] or
            result == [{"price": 19.95, "color": "red"}, [{"category": "reference", "price": 8.95, "author": "Nigel Rees", "title": "Sayings of the Century"}, {"category": "fiction", "price": 12.99, "author": "Evelyn Waugh", "title": "Sword of Honour"}, {"category": "fiction", "isbn": "0-553-21311-3", "price": 8.99, "author": "Herman Melville", "title": "Moby Dick"}, {"category": "fiction", "isbn": "0-395-19395-8", "price": 22.99, "author": "J. R. R. Tolkien", "title": "The Lord of the Rings"}]]):
            success = True
        self.assertEqual(success, True)

    def test_goessner04(self):
        print("The price of everything in the store.")
        result = apply_jsonpath(goessner, "$.store..price")
        print(result)
        success = False
        # The "bicycle" and "book" results could be returned in either order
        # in the result list so need to check both
        if (result == [8.95, 12.99, 8.99, 22.99, 19.95] or
            result == [19.95, 8.95, 12.99, 8.99, 22.99]):
            success = True
        self.assertEqual(success, True)

    def test_goessner05(self):
        print("The third book.")
        result = apply_jsonpath(goessner, "$..book[2]")
        print(result)
        self.assertEqual(result, {"title": "Moby Dick", "category": "fiction", "author": "Herman Melville", "price": 8.99, "isbn": "0-553-21311-3"})

    def test_goessner06(self):
        print("The last book in the order.")
        result = apply_jsonpath(goessner, "$..book[(@.length-1)]")
        print(result)
        self.assertEqual(result, {"category": "fiction", "price": 22.99, "author": "J. R. R. Tolkien", "isbn": "0-395-19395-8", "title": "The Lord of the Rings"})

        # Or alternatively using slicing. Note that in this case a single item
        # array containing the object is returned from the slice operator.
        result = apply_jsonpath(goessner, "$..book[-1:]")
        print(result)
        self.assertEqual(result, [{"category": "fiction", "price": 22.99, "author": "J. R. R. Tolkien", "isbn": "0-395-19395-8", "title": "The Lord of the Rings"}])

    def test_goessner07(self):
        print("The first two books.")
        result = apply_jsonpath(goessner, "$..book[0,1]")
        print(result)
        self.assertEqual(result, [{"title": "Sayings of the Century", "author": "Nigel Rees", "category": "reference", "price": 8.95}, {"title": "Sword of Honour", "author": "Evelyn Waugh", "category": "fiction", "price": 12.99}])
        # Or alternatively using different syntax/filter
        result = apply_jsonpath(goessner, "$..book[:2]")
        print(result)
        self.assertEqual(result, [{"title": "Sayings of the Century", "author": "Nigel Rees", "category": "reference", "price": 8.95}, {"title": "Sword of Honour", "author": "Evelyn Waugh", "category": "fiction", "price": 12.99}])

    def test_goessner08(self):
        print("Filter all books with isbn number.")
        result = apply_jsonpath(goessner, "$..book[?(@.isbn)]")
        self.assertEqual(result, [{"isbn": "0-553-21311-3", "price": 8.99, "title": "Moby Dick", "author": "Herman Melville", "category": "fiction"}, {"isbn": "0-395-19395-8", "price": 22.99, "title": "The Lord of the Rings", "author": "J. R. R. Tolkien", "category": "fiction"}])

    def test_goessner09(self):
        print("Filter all books cheaper than 10")
        result = apply_jsonpath(goessner, "$..book[?(@.price<10)]")
        print(result)
        self.assertEqual(result, [{"title": "Sayings of the Century", "author": "Nigel Rees", "price": 8.95, "category": "reference"}, {"isbn": "0-553-21311-3", "title": "Moby Dick", "author": "Herman Melville", "price": 8.99, "category": "fiction"}])

    def test_goessner10(self):
        print("All members of JSON structure.")
        result = apply_jsonpath(goessner, "$..*")
        print(result)
        """
        TODO equality test. There are lots of permutations of correct result
        due to unordered nature of dict (at least for some python versions) so
        the returned list will potentially reflect that ordering. Comparing
        correctly is therefore long and rather tedious.
        """

    def test_index_out_of_bounds(self):
        print("Apply path with index out of bounds")

        result = None
        try:
            result = apply_jsonpath(simple_array, "$.inputs[2]")
        except PathMatchFailure as e:
            print("Correctly caught exception PathMatchFailure")

        self.assertEqual(result, None)

        try:
            result = apply_jsonpath(empty_array, "$.inputs[0]")
        except PathMatchFailure as e:
            print("Correctly caught exception PathMatchFailure")

        self.assertEqual(result, None)

if __name__ == "__main__":
    unittest.main()




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 id_then_unzip.py
#
"""
Note that this test makes use of an enhancement whereby the state machine
Definition can be passed as well as its Id the context. This means that the
create-state-machine API call does not need to be explicitly called prior to
running this test as passing the Definition in the context implicitly performs
the create-state-machine action.
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.amqp_0_9_1_messaging import Connection, Message
from asl_workflow_engine.messaging_exceptions import *

ASL = """{
  "Comment": "A Simple step function to perform MIME ID followed by untar",
  "StartAt": "MimeID",
  "States": {
    "MimeID": {
      "Type": "Task",
      "Resource": "arn:aws:rpcmessage:local::function:mime-id",
      "ResultPath": null,
      "Next": "BsdtarUnzip"
    },
    "BsdtarUnzip": {
      "Type": "Task",
      "Resource": "arn:aws:rpcmessage:local::function:bsdtar-unzip",
      "End": true
    }
  }
}"""


"""
See https://stackoverflow.com/questions/2150739/iso-time-iso-8601-in-python
For info on creating ISO 8601 time format

The application context is described in the AWS documentation:
https://docs.aws.amazon.com/step-functions/latest/dg/input-output-contextobject.html 

{
    "Execution": {
        "Id": <String>,
        "Input": <Object>,
        "StartTime": <String Format: ISO 8601>
    },
    "State": {
        "EnteredTime": <String Format: ISO 8601>,
        "Name": <String>,
        "RetryCount": <Number>
    },
    "StateMachine": {
        "Id": <String>,
        "Definition": <Object representing ASL state machine>
    },
    "Task": {
        "Token": <String>
    }
}

The most important paths for state traversal are:
$$.State.Name = the current state
$$.StateMachine.Definition = (optional) contains the complete ASL state machine
$$.StateMachine.Id = a unique reference to an ASL state machine
"""
context = '{"StateMachine": {"Id": "arn:aws:states:local:0123456789:stateMachine:id_then_unzip", "Definition": ' + ASL + '}}'

items = ['{"data": {"zipfile": "s3://37199-dev/CFX/input-data/akismet.2.5.3.zip", "destination": "s3://37199-dev/CFX/processed-data"}, "context": ' + context + '}']

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name='id_then_unzip')

    # Connect to event queue and send items.
    connection = Connection("amqp://localhost:5672?connection_attempts=20&retry_delay=10&heartbeat=0")
    try:
        connection.open()
        session = connection.session()
        producer = session.producer("asl_workflow_events") # event queue
        for item in items:
            """
            Setting content_type isn't necessary for correct operation,
            however it is the correct thing to do:
            https://www.ietf.org/rfc/rfc4627.txt.
            """
            producer.send(Message(item, content_type="application/json"))
        connection.close();
    # Could catch MessagingError if we don't want to handle these separately
    except ConnectionError as e:
        self.logger.error(e)
    except SessionError as e:
        self.logger.error(e)
    except ConsumerError as e:
        self.logger.error(e)
    except ProducerError as e:
        self.logger.error(e)




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3


import asyncio, importlib
from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.messaging_exceptions import *

try:  # Attempt to use ujson if available https://pypi.org/project/ujson/
    import ujson as json
except:  # Fall back to standard library json
    import json

class EventDispatcher(object):
    def __init__(self, state_engine, config):
        """
        :param logger: The Workflow Engine logger
        :type logger: logger
        :param config: Configuration dictionary
        :type config: dict
        """
        self.logger = init_logging(log_name="asl_workflow_engine")
        self.logger.info("Creating EventDispatcher, using {} JSON parser".format(json.__name__))

        self.session = None  # Make session an attribute so it's visible from REST API

        self.queue_config = config["event_queue"]  # TODO Handle missing config
        self.notifier_config = config["notifier"]  # TODO Handle missing config
        # TODO validate that config contains the keys we need.

        self.queue_name = self.queue_config.get("queue_name", "asl_workflow_events")

        instance_id = self.queue_config.get("instance_id", "")
        self.instance_queue_name = self.queue_name + "-" + instance_id

        # Get consumer capacities from config as numbers or numeric strings
        capacity = self.queue_config.get("shared_event_consumer_capacity", 1000)
        self.shared_event_consumer_capacity = int(float(capacity))

        capacity = self.queue_config.get("instance_event_consumer_capacity", 1000)
        self.instance_event_consumer_capacity = int(float(capacity))

        """
        Create an association with the state engine and give that a reference
        back to this event dispatcher so that it can publish events and make
        use of the set_timeout time scheduler.
        """
        self.state_engine = state_engine
        self.state_engine.event_dispatcher = self

        """
        Incoming messages should only be acknowledged when they have been
        fully processed by the StateEngine to allow the messaging fabric to
        redeliver the message upon a failure. Because processing might take
        some time due to waiting for Task responses and the Wait and Parallel
        states we must retain any unacknowledged messages. The message count
        is a simple one-up number used as a key.
        """
        self.unacknowledged_messages = {}
        self.shared_event_consumer_unack_count = 0
        self.instance_event_consumer_unack_count = 0
        self.message_count = 0

        """
        """
        self.heartbeat_count = 0

        """
        Connection Factory for the event queue. The idea is to eventually
        allow the ASL workflow engine to connect to alternative event queue
        implementations in order to allow maximum flexibility.
        """
        name = (
            self.queue_config.get("queue_type", "AMQP-0.9.1")
            .lower()
            .replace("-", "_")
            .replace(".", "_")
        )

        if name.endswith("_asyncio"):
            self.name = "asl_workflow_engine." + name[:-8] + "_messaging_asyncio"
        else:
            self.name = "asl_workflow_engine." + name + "_messaging"

        self.logger.info("Loading messaging module {}".format(self.name))

        # Load the module whose name is derived from the specified queue_type.
        try:
            print('name',self.name)
            messaging = importlib.import_module("asl_workflow_engine.amqp_0_9_1_messaging")
            globals()["Connection"] = messaging.Connection
            globals()["Message"] = messaging.Message
        except ImportError as e:
            self.logger.error(e)
            sys.exit(1)

    
    def heartbeat(self):
        print("**** EventDispatcher heartbeat ****")
        #self.state_engine.heartbeat()
        """
        self.heartbeat_count += 1
        if self.heartbeat_count % 60 == 0:
            print("self.instance_event_consumer_unack_count")
            print(self.instance_event_consumer_unack_count)

            print("len(self.unacknowledged_messages)")
            print(len(self.unacknowledged_messages))
        """
        self.set_timeout(self.heartbeat, 1000)
    

    def start(self):
        # Connect to event queue and start the main event loop.
        # TODO This code will connect on broker startup, but need to add code to
        # reconnect for cases where the broker fails and then restarts.
        connection = Connection(self.queue_config["connection_url"])
        """
        TODO Passing connection.set_timeout here is kind of ugly but I can't
        think of a nicer way yet. The issue is that we want the state engine
        to be able to support timeouts, however that is being called from
        Pika's event loop, which isn't thread-safe and we want asynchronous
        timeouts not blocking timeouts as when in an ASL Wait state we still
        want to be able to process other events that might be available on
        the event queue. This (hopefully temporary) approach is Pika specific.
        """
        self.set_timeout = connection.set_timeout
        self.clear_timeout = connection.clear_timeout

        try:
            connection.open()
            self.session = connection.session()

            """
            Share messaging session with state_engine.task_dispatcher. This
            is to allow rpcmessage invocations that share the same message
            fabric instance as the event queue to reuse connections etc.
            """
            self.state_engine.task_dispatcher.start(self.session)

            """
            The event_queue_producer is used to publish events corresponding
            to state transitions. This can publish to both the asl_workflow_events
            queue that is shared by every asl_workflow_engine instance and also
            the per-instance queues of each instance. The Message subject is
            used to select which queue the Message should be published to.
            """
            self.event_queue_producer = self.session.producer(self.queue_name)

            """
            The topic_producer specifies the topic that notification events
            should be published to. Notification events are analogous to AWS
            CloudWatch events and the JSON format of the notification Messages
            is the same as that used by CloudWatch.
            """
            self.topic_producer = self.session.producer(self.notifier_config["topic"])

            """
            The asl_workflow_events queue is a shared event queue, that is to
            say every asl_workflow_engine instance receives events from this
            queue. This is used to publish StartExecution events to and is non-
            exclusive so multiple asl_workflow_engine instances can consume from
            it and will thus load-balance executions across multiple instances.
            """
            shared_queue = self.queue_name + '; {"node": {"durable": true}}'
            shared_event_consumer = self.session.consumer(shared_queue)
            # Enable consumer prefetch
            self.logger.info("Setting shared_event_consumer.capacity to {}".format(
                self.shared_event_consumer_capacity)
            )
            shared_event_consumer.capacity = self.shared_event_consumer_capacity
            shared_event_consumer.set_message_listener(self.dispatch)

            """
            the instance_event_consumer is an event queue that is set up for
            each asl_workflow_engine instance. This should be an exclusive
            queue such that only a single instance can consume from it. This
            queue is used for the remainder of the events for each execution.
            The idea of having an event queue per instance is because most
            messaging implementations scale across queues so simply having lots
            of consumers on a single queue is fine if the bottleneck is due to
            the consumer performance, but if the bottleneck is due to limits
            of the messaging fabric then it won't help. Another reason for a
            per instance queue is because with the Parallel and Map states we
            would like each branch to notify the same instance when complete.
            """
            instance_queue = self.instance_queue_name + '; {"node": {"durable": true}, "link": {"x-subscribe": {"exclusive": true}}}'
            instance_event_consumer = self.session.consumer(instance_queue)
            # Enable consumer prefetch
            self.logger.info("Setting instance_event_consumer.capacity to {}".format(
                self.instance_event_consumer_capacity)
            )
            instance_event_consumer.capacity = self.instance_event_consumer_capacity
            instance_event_consumer.set_message_listener(self.dispatch)

            """
            Start a periodic "heartbeat". The idea of this is that sometimes
            "bad things happen", for example if a service goes down we might
            be sat stuck in a Task state. Now things *should* carry on when the
            service is restarted, but for that to work we'd need to ensure
            redelivery of unacknowledged messages via recover call. Similarly
            in the TaskDispatcher we might have message requests for which no
            responses have been received and those are likely to require reaping
            for cases where no explicit timeout has been added to the State.
            """
            #self.set_timeout(self.heartbeat, 1000)

            connection.start()  # Blocks until event loop exits.
        except MessagingError as e:
            self.logger.error(e)
            sys.exit(1)

        connection.close()

    # asyncio version of the start() method above
    async def start_asyncio(self):
        # Connect to event queue and start the main event loop.
        # TODO This code will connect on broker startup, but need to add code to
        # reconnect for cases where the broker fails and then restarts.
        connection = Connection(self.queue_config["connection_url"])
        """
        TODO Passing connection.set_timeout here is kind of ugly but I can't
        think of a nicer way yet. The issue is that we want the state engine
        to be able to support timeouts, however that is being called from
        Pika's event loop, which isn't thread-safe and we want asynchronous
        timeouts not blocking timeouts as when in an ASL Wait state we still
        want to be able to process other events that might be available on
        the event queue. This (hopefully temporary) approach is Pika specific.
        """
        self.set_timeout = connection.set_timeout
        self.clear_timeout = connection.clear_timeout

        try:
            await connection.open()
            self.session = await connection.session()

            """
            Share messaging session with state_engine.task_dispatcher. This
            is to allow rpcmessage invocations that share the same message
            fabric instance as the event queue to reuse connections etc.
            """
            await self.state_engine.task_dispatcher.start_asyncio(self.session)

            """
            The event_queue_producer is used to publish events corresponding
            to state transitions. This can publish to both the asl_workflow_events
            queue that is shared by every asl_workflow_engine instance and also
            the per-instance queues of each instance. The Message subject is
            used to select which queue the Message should be published to.
            """
            self.event_queue_producer = await self.session.producer(self.queue_name)

            """
            The topic_producer specifies the topic that notification events
            should be published to. Notification events are analogous to AWS
            CloudWatch events and the JSON format of the notification Messages
            is the same as that used by CloudWatch.
            """
            self.topic_producer = await self.session.producer(self.notifier_config["topic"])

            """
            The asl_workflow_events queue is a shared event queue, that is to
            say every asl_workflow_engine instance receives events from this
            queue. This is used to publish StartExecution events to and is non-
            exclusive so multiple asl_workflow_engine instances can consume from
            it and will thus load-balance executions across multiple instances.
            """
            shared_queue = self.queue_name + '; {"node": {"durable": true}}'
            shared_event_consumer = await self.session.consumer(shared_queue)
            # Enable consumer prefetch
            self.logger.info("Setting shared_event_consumer.capacity to {}".format(
                self.shared_event_consumer_capacity)
            )
            shared_event_consumer.capacity = self.shared_event_consumer_capacity
            await shared_event_consumer.set_message_listener(self.dispatch)

            """
            the instance_event_consumer is an event queue that is set up for
            each asl_workflow_engine instance. This should be an exclusive
            queue such that only a single instance can consume from it. This
            queue is used for the remainder of the events for each execution.
            The idea of having an event queue per instance is because most
            messaging implementations scale across queues so simply having lots
            of consumers on a single queue is fine if the bottleneck is due to
            the consumer performance, but if the bottleneck is due to limits
            of the messaging fabric then it won't help. Another reason for a
            per instance queue is because with the Parallel and Map states we
            would like each branch to notify the same instance when complete.
            """
            instance_queue = self.instance_queue_name + '; {"node": {"durable": true}, "link": {"x-subscribe": {"exclusive": true}}}'
            instance_event_consumer = await self.session.consumer(instance_queue)
            # Enable consumer prefetch
            self.logger.info("Setting instance_event_consumer.capacity to {}".format(
                self.instance_event_consumer_capacity)
            )
            instance_event_consumer.capacity = self.instance_event_consumer_capacity
            await instance_event_consumer.set_message_listener(self.dispatch)

            """
            Start a periodic "heartbeat". The idea of this is that sometimes
            "bad things happen", for example if a service goes down we might
            be sat stuck in a Task state. Now things *should* carry on when the
            service is restarted, but for that to work we'd need to ensure
            redelivery of unacknowledged messages via recover call. Similarly
            in the TaskDispatcher we might have message requests for which no
            responses have been received and those are likely to require reaping
            for cases where no explicit timeout has been added to the State.
            """
            #self.set_timeout(self.heartbeat, 1000)


            def on_close_callback(ch, e):
                code = type(e).__name__
                messsage = f"Unexpected channel closure triggered an on_close_callback with exception: {code}: {str(e)}. Terminating!"
                self.logger.error(messsage)
                sys.exit(1)

            self.session.channel.add_on_close_callback(on_close_callback)

            await connection.start()  # Blocks until event loop exits.
        except MessagingError as e:
            self.logger.error(e)
            sys.exit(1)

        connection.close()

    def dispatch(self, message):
        """
        Dispatch the body of the message received from event queue hosted on the
        underlying messaging fabric. This method is mainly here to abstract some
        of the implementation details from the state engine, which should be
        somewhat agnostic of the event queue implementation.

        The message body *should* be a JSON string, but we trap any exception
        that may be thrown during conversion and simply dump the message and
        log the failure. Note that some messaging systems (such as Pika)
        incorrectly, IMHO, encode strings as *opaque binary* on the messaging
        fabric rather than string. This can lead to errors when doing things
        like JSON decoding as json.loads(message.body) works fine in Python 3.6
        but does not in earlier versions. This is kind of irksome given that
        Python has a string type and supports introspection, but there you go.
        """
        try:
            item = json.loads(message.body.decode("utf8"))
            self.unacknowledged_messages[self.message_count] = message

            if message.subject == self.queue_name:
                self.shared_event_consumer_unack_count += 1
                #print("self.shared_event_consumer_unack_count")
                #print(self.shared_event_consumer_unack_count)
            else:
                self.instance_event_consumer_unack_count += 1
                #print("self.instance_event_consumer_unack_count")
                #print(self.instance_event_consumer_unack_count)

            self.state_engine.notify(item, self.message_count)
            self.message_count += 1
        except ValueError as e:
            self.logger.error(
                "Message {} does not contain valid JSON".format(message.body)
            )
            message.acknowledge(multiple=False)
        except Exception as e:
            """
            If state_engine.notify bombs out with an exception it is likely to
            be due to invalid data or the ASL not being handled correctly.
            It's hard to know the best course of action, but for now catch the
            Exception, log error then acknowledge the "poison" message to
            prevent it from being endlessly redelivered.
            """
            self.logger.error(
                "Message {} caused the exception: {}:{} - dropping the message!".format(
                    message.body, type(e).__name__, str(e)
                )
            )
            message.acknowledge(multiple=False)

    def acknowledge(self, id):
        """
        Look up the message with the given id in the unacknowledged_messages
        dictionary and acknowledge it, then remove the message from the
        dictionary. See the comment for unacknowledged_messages in constructor.
        """
        message = self.unacknowledged_messages[id]
        message.acknowledge(multiple=False)
        del self.unacknowledged_messages[id]

        if message.subject == self.queue_name:
            self.shared_event_consumer_unack_count -= 1
            #print("self.shared_event_consumer_unack_count")
            #print(self.shared_event_consumer_unack_count)
        else:
            self.instance_event_consumer_unack_count -= 1
            #print("self.instance_event_consumer_unack_count")
            #print(self.instance_event_consumer_unack_count)

    def publish(self, item, threadsafe=False, start_execution=False):
        """
        Publish the supplied item to the event queue hosted on the underlying
        messaging fabric. This method is mainly here to abstract some of the
        implementation details from the state engine.

        The start_execution field is used to select between publishing to the
        event queue that is shared by all instances of asl_workflow_engine and
        the per instance queue.

        Setting content_type to application/json isn't necessary for correct
        operation, however it is the correct thing to do:
        https://www.ietf.org/rfc/rfc4627.txt.
        """
        if start_execution:
            subject = self.queue_name
        else:
            subject = self.instance_queue_name

        message = Message(json.dumps(item), content_type="application/json")
        message.subject = subject  # Selects the queue to publish to
        self.event_queue_producer.send(message, threadsafe)

    def broadcast(self, subject, item, carrier_properties=None):
        """
        Broadcast the supplied item to the topic hosted on the underlying
        messaging fabric. This method is mainly here to abstract some of the
        implementation details from the state engine.

        Setting content_type to application/json isn't necessary for correct
        operation, however it is the correct thing to do:
        https://www.ietf.org/rfc/rfc4627.txt.
        """
        message = Message(
            json.dumps(item),
            content_type="application/json",
            properties=carrier_properties
        )
        message.subject = subject
        self.topic_producer.send(message)




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 simple_state_machine1.py
#
"""
This test is the rough equivalent of running the following on the AWS CLI
with item 1 to 4 differing by the --input below.

Note that this test makes use of an enhancement whereby the state machine
Definition can be passed as well as its Id the context. This means that the
create-state-machine API call does not need to be explicitly called prior to
running this test as passing the Definition in the context implicitly performs
the create-state-machine action.

aws stepfunctions --endpoint http://localhost:4584 start-execution --state-machine-arn $STATE_MACHINE_ARN --input '{"lambda":"Success"}' --name success-execution

# InternalErrorNotHandled lambda
aws stepfunctions --endpoint http://localhost:4584 start-execution --state-machine-arn $STATE_MACHINE_ARN --input '{"lambda":"InternalErrorNotHandled"}' --name internal-error-not-handled-execution

# InternalErrorHandled lambda
aws stepfunctions --endpoint http://localhost:4584 start-execution --state-machine-arn $STATE_MACHINE_ARN --input '{"lambda":"InternalErrorHandled"}' --name internal-error-handled-execution

# Timeout lambda
aws stepfunctions --endpoint http://localhost:4584 start-execution --state-machine-arn $STATE_MACHINE_ARN --input '{"lambda":"Timeout"}' --name timeout-execution
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.amqp_0_9_1_messaging import Connection, Message
from asl_workflow_engine.messaging_exceptions import *

ASL = """{
    "Comment": "Test Step Function",
    "StartAt": "StartState",
    "States": {
        "StartState": {
            "Type": "Pass",
            "Next": "ChoiceState"
        },
        "ChoiceState": {
            "Type": "Choice",
            "Choices": [
                {
                    "Variable": "$.lambda",
                    "StringEquals": "InternalErrorNotHandled",
                    "Next": "InternalErrorNotHandledLambda"
                },
                {
                    "Variable": "$.lambda",
                    "StringEquals": "InternalErrorHandled",
                    "Next": "InternalErrorHandledLambda"
                },
                {
                    "Variable": "$.lambda",
                    "StringEquals": "Success",
                    "Next": "SuccessLambda"
                },
                {
                    "Variable": "$.lambda",
                    "StringEquals": "Timeout",
                    "Next": "TimeoutLambda"
                }
            ],
            "Default": "FailState"
        },
        "FailState": {
            "Type": "Fail",
            "Error": "NoLambdaError",
            "Cause": "No Matches!"
        },
        "SuccessLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:SuccessLambda",
            "Next": "WaitState"
        },
        "InternalErrorNotHandledLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:InternalErrorNotHandledLambda",
            "Next": "EndState"
        },
        "InternalErrorHandledLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:InternalErrorHandledLambda",
            "Next": "EndState"
        },
        "TimeoutLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:TimeoutLambda",
            "Next": "EndState"
        },
        "EndState": {
            "Type": "Pass",
            "End": true
        },
        "WaitState": {
            "Type": "Wait",
            "Seconds":10,
            "Next": "EndState"
        }
    }
}"""


"""
The application context is described in the AWS documentation:
https://docs.aws.amazon.com/step-functions/latest/dg/input-output-contextobject.html 

{
    "Execution": {
        "Id": <String>,
        "Input": <Object>,
        "StartTime": <String Format: ISO 8601>
    },
    "State": {
        "EnteredTime": <String Format: ISO 8601>,
        "Name": <String>,
        "RetryCount": <Number>
    },
    "StateMachine": {
        "Id": <String>,
        "Definition": <Object representing ASL state machine>
    },
    "Task": {
        "Token": <String>
    }
}

The most important paths for state traversal are:
$$.State.Name = the current state
$$.StateMachine.Definition = (optional) contains the complete ASL state machine
$$.StateMachine.Id = a unique reference to an ASL state machine
"""
context = '{"StateMachine": {"Id": "arn:aws:states:local:0123456789:stateMachine:simple_state_machine", "Definition": ' + ASL + '}}'

items = ['{"data": {"lambda":"Success"}, "context": ' + context + '}',
         '{"data": {"lambda":"InternalErrorNotHandled"}, "context": ' + context + '}',
         '{"data": {"lambda":"InternalErrorHandled"}, "context": ' + context + '}',
         '{"data": {"lambda":"Timeout"}, "context": ' + context + '}']

items = ['{"data": {"lambda":"Success"}, "context": ' + context + '}']
#items = ['{"data": {"lambda":"InternalErrorNotHandled"}, "context": ' + context + '}']
#items = ['{"data": {"lambda":"InternalErrorHandled"}, "context": ' + context + '}']
#items = ['{"data": {"lambda":"Timeout"}, "context": ' + context + '}']

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name='simple_state_machine1')

    # Connect to event queue and send items.
    connection = Connection("amqp://localhost:5672?connection_attempts=20&retry_delay=10&heartbeat=0")
    try:
        connection.open()
        session = connection.session()
        producer = session.producer("asl_workflow_events") # event queue
        for item in items:
            """
            Setting content_type isn't necessary for correct operation,
            however it is the correct thing to do:
            https://www.ietf.org/rfc/rfc4627.txt.
            """
            producer.send(Message(item, content_type="application/json"))
        connection.close()
    # Could catch MessagingError if we don't want to handle these separately
    except ConnectionError as e:
        logger.error(e)
    except SessionError as e:
        logger.error(e)
    except ConsumerError as e:
        logger.error(e)
    except ProducerError as e:
        logger.error(e)




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 iterate2.py
#
"""
This test illustrates the case of passing an array of data to a Step Function
and iterating through the array, triggering execution of a child Step Function
for each item in the array. This test is mostly  intended to provide quite a bit
of "stress" to the ASL engine as it can result in rather large JSON objects as
inputs to the state machine, which will tend to make JSONPath processing the
bottleneck, and lots of items in the input JSON array will also result in lots
of state transitions for each execution.


This test uses the AWS python SDK boto3 to access our local ASL Workflow Engine
https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/stepfunctions.html
https://docs.aws.amazon.com/code-samples/latest/catalog/code-catalog-python-example_code-stepfunctions.html
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import boto3, time
from botocore.exceptions import ClientError

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer

iterate2_ASL = """{
    "Comment": "State Machine to illustrate iterating an array and passing each array item to a child Step Function",
    "StartAt": "HasIndexZero",
    "States": {
        "HasIndexZero": {
            "Type": "Choice",
            "Choices": [
                {
                    "Variable": "$[0]",
                    "IsPresent": true,
                    "Next": "StepFunctionLauncher"
                }
            ],
            "Default": "EndState"
        },
        "StepFunctionLauncher": {
            "Comment": "Launch the child Step Function. In the Parameters section we extract the first item from the array passed as input to the state and use that as the input to the child Step Function. The ResultPath of null causes the input to be passed to the output.",
            "Type": "Task",
            "Resource": "arn:aws:states:local:0123456789:states:startExecution",
            "Parameters": {  
                "Input.$": "$[0]",
                "StateMachineArn": "arn:aws:states:local:0123456789:stateMachine:child_state_machine"
            },
            "ResultPath": null,
            "Next": "HasMultipleElements"
        },
        "HasMultipleElements": {
            "Type": "Choice",
            "Choices": [
                {
                    "Variable": "$[1:]",
                    "IsPresent": true,
                    "Next": "RemoveFirstItemFromArray"
                }
            ],
            "Default": "EndState"
        },
        "RemoveFirstItemFromArray": {
            "Type": "Pass",
            "InputPath": "$[1:]",
            "Next": "StepFunctionLauncher"
        },
        "EndState": {
            "Type": "Pass",
            "End": true,
            "Parameters": []
        }
    }
}"""

child_ASL = """{
    "Comment": "Trivial Child Step Function",
    "StartAt": "StartState",
    "States": {
        "StartState": {
            "Type": "Pass",
            "End": true
        }
    }
}"""

array_length = 1000
item = "[0" + "".join(", " + str(i) for i in range(1, array_length)) + "]"
#items = [item]
items = [item, item, item, item, item, item, item, item, item, item, item, item, item, item, item, item, item, item, item, item]

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name='iterate2')

    # Initialising OpenTracing. It's important to do this before the boto3.client
    # call as create_tracer "patches" boto3 to add the OpenTracing hooks.
    create_tracer("iterate2", {"implementation": "Jaeger"})

    # Initialise the boto3 client setting the endpoint_url to our local
    # ASL Workflow Engine
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session.client
    sfn = boto3.client("stepfunctions", endpoint_url="http://localhost:4584")
    iterate2_state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:iterate2_state_machine"
    child_state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:child_state_machine"

    def create_state_machines():
        # Create state machines using a dummy roleArn. If it already exists an
        # exception will be thrown, we ignore that but raise other exceptions.
        try:
            response = sfn.create_state_machine(
                name="iterate2_state_machine", definition=iterate2_ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )

            response = sfn.create_state_machine(
                name="child_state_machine", definition=child_ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )
        except sfn.exceptions.StateMachineAlreadyExists as e:
            #print(e.response)

            response = sfn.update_state_machine(
                stateMachineArn=iterate2_state_machine_arn,
                definition=iterate2_ASL
            )

            response = sfn.update_state_machine(
                stateMachineArn=child_state_machine_arn,
                definition=child_ASL
            )
        except ClientError as e:
            logger.error(e)


    create_state_machines()

    # Loop through items invoking a new state machine execution for each item
    for item in items:
        try:
            response = sfn.start_execution(
                stateMachineArn=iterate2_state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except sfn.exceptions.StateMachineDoesNotExist as e:
            logger.info(e)

            create_state_machines()
            response = sfn.start_execution(
                stateMachineArn=iterate2_state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except ClientError as e:
            logger.error(e)

    time.sleep(1)  # Give OpenTracing a chance to flush buffer before exiting




import sys
import unittest
from unittest import mock
import asyncio

assert sys.version_info >= (3, 0) # Bomb out if not running Python3

from asl_workflow_engine.rest_api import RestAPI as sync_RestAPI
from asl_workflow_engine.rest_api_asyncio import RestAPI as async_RestAPI

class health_checker(unittest.TestCase):

    # Setup to mock necessary components
    def setUp(self):
        self.event_dispatcher = mock.MagicMock()
        self.session = mock.MagicMock()
        # Default to True, changed in false tests
        self.session.is_open = mock.MagicMock(return_value=True)
        self.event_dispatcher.session = self.session

        # Default to sync RestAPI, changed in async tests
        self.rest_api = sync_RestAPI(mock.MagicMock(), self.event_dispatcher, {})

    # Test for sync session open 
    #   No value changed
    def test_session_open(self):
        app = self.rest_api.create_app()
        test_rest_client = app.test_client()
        Response = test_rest_client.get("/health")
        self.assertEqual(Response.status_code, 200)

    # Test for sync session closed
    #   Session channel open value changed
    def test_session_closed(self):
        self.session.is_open = mock.MagicMock(return_value=False)
        app = self.rest_api.create_app()
        test_rest_client = app.test_client()
        Response = test_rest_client.get("/health")
        self.assertEqual(Response.status_code, 503)        

    # Test for async session closed 
    #   Sesssion channel open value changed
    #   API mock changed to async
    def test_async_session_closed(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self.session.is_open = mock.MagicMock(return_value=False)
        self.rest_api = async_RestAPI(mock.MagicMock(), self.event_dispatcher, {})
        app = self.rest_api.create_app()
        test_rest_client = app.test_client()
        Response = asyncio.run(test_rest_client.get("/health"))
        self.assertEqual(Response.status_code, 503) 


    # Test for async session open
    #   API mock changed to async
    def test_async_session_open(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self.rest_api = async_RestAPI(mock.MagicMock(), self.event_dispatcher, {})
        app = self.rest_api.create_app()
        test_rest_client = app.test_client()
        Response = asyncio.run(test_rest_client.get("/health"))
        self.assertEqual(Response.status_code, 200) 

    if __name__ == '__main__':
        unittest.main()


#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 subscribe_to_notifications_asyncio.py
#

"""
This test application acts as a simple "NotificationHandler" the idea being that
it subscribes to notification events from the asl_workflow_engine when the
State Machine starts or ends (e.g. RUNNING, SUCCEEDED or FAILED etc.)
The intention is to act rather like AWS CloudWatch events:
https://docs.aws.amazon.com/step-functions/latest/dg/cw-events.html
https://docs.aws.amazon.com/AmazonCloudWatch/latest/events/CloudWatchEventsandEventPatterns.html

The events are published to:
asl_workflow_engine/<state_machine_arn>.<status>

That is to say the asl_workflow_engine topic exchange with a subject comprising
the state_machine_arn and the status separated by a dot. e.g. 
asl_workflow_engine/arn:aws:states:local:0123456789:stateMachine:simple_state_machine.SUCCEEDED

Subscribe to specific events using the full subject or groups of event using
wildcards, for example:
asl_workflow_engine/arn:aws:states:local:0123456789:stateMachine:simple_state_machine.*
Subscribes to all notification events published for a given state machine.

asl_workflow_engine/#
Subscribes to all events from the asl_workflow_engine

The format of the message body is as described in
https://docs.aws.amazon.com/step-functions/latest/dg/cw-events.html
the "detail" field gives access to the "executionArn", "stateMachineArn",
"input", "output" and other relevant information.
"""

# Just an aide-memoire for state machine and execution ARN formats
# state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:simple_state_machine"
# name = str(uuid.uuid4())
# execution_arn = "arn:aws:states:local:0123456789:execution:simple_state_machine:" + name

import sys
assert sys.version_info >= (3, 6) # Bomb out if not running Python3.6

import asyncio, json, time

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.amqp_0_9_1_messaging_asyncio import Connection, Message
from asl_workflow_engine.messaging_exceptions import *

class NotificationHandler(object):
    def __init__(self, state_machines):
        # Initialise logger
        self.logger = init_logging(log_name="subscribe_to_notifications")
        self.state_machines = state_machines

    async def run(self):
        connection = Connection("amqp://localhost:5672?connection_attempts=20&retry_delay=10&heartbeat=0")
        try:
            await connection.open()
            session = await connection.session()

            for s in self.state_machines:
                """
                Use nested function so we can capture the subscription value s.
                Now closures look at the value in the surrounding scope at the
                *time the function is called* so just using a simple nested
                def handler(message): function would always give the last value
                of the loop. The following creates a scoped closure where we
                pass in the subscription value in the loop and return the actual
                handler as a closure that has the subscription in its scope.
                https://stackoverflow.com/questions/12423614/local-variables-in-nested-functions
                """
                def scoped_handler(subscription):                
                    def handler(message):
                        print("Message received for subscription:")
                        print(subscription)
                        print(message)
                        print()
                        message.acknowledge() # Acknowledges the original request
                    return handler

                # Example topic subscription with options to create (declare)
                # the topic exchange if it doesn't already exist.
                #consumer = await session.consumer("asl_workflow_engine/" + s + '; {"node": {"x-declare": {"exchange": "asl_workflow_engine", "exchange-type": "topic", "durable": true}}}')

                # Simple topic subscription - will fail if the specified topic
                # exchange doesn't already exist.
                consumer = await session.consumer("asl_workflow_engine/" + s)
                consumer.capacity = 100; # Enable consumer prefetch
                await consumer.set_message_listener(scoped_handler(s))


            await connection.start(); # Blocks until event loop exits.

        except MessagingError as e:
            self.logger.error(e)

        connection.close();

if __name__ == '__main__':
    # Subscribe to notification events from these State Machines
    state_machines = [
        "arn:aws:states:local:0123456789:stateMachine:simple_state_machine.*",  
        "arn:aws:states:local:0123456789:stateMachine:caller_state_machine.*",
        "arn:aws:states:local:0123456789:stateMachine:parallel1.*",
        "arn:aws:states:local:0123456789:stateMachine:map1.*",
        "arn:aws:states:local:0123456789:stateMachine:iterate2_state_machine.*",
    ]
    notification_handler = NotificationHandler(state_machines)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(notification_handler.run())
    loop.close()




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
The "official" prometheus Python client provides metrics for Standard Exports
such as gc, cpu and memory, but aioprometheus does not provide these by default. 
This module provides equivalent system metrics for aioprometheus.
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3

import gc, os, platform
from aioprometheus import Counter, Gauge

try:
    import resource
    _PAGESIZE = resource.getpagesize()
except ImportError:
    # Not Unix
    _PAGESIZE = 4096


class SystemMetrics(object):
    """
    This class provides aioprometheus with equivalent system level metrics
    to those found in the "official" prometheus Python client.
    https://github.com/prometheus/client_python/blob/master/prometheus_client/process_collector.py
    https://github.com/prometheus/client_python/blob/master/prometheus_client/platform_collector.py
    https://github.com/prometheus/client_python/blob/master/prometheus_client/gc_collector.py
    """
    def __init__(self, namespace=""):
        if namespace:
            ns = namespace + '_'
        else:
            ns = ""

        self._ticks = 100.0
        try:
            self._ticks = os.sysconf('SC_CLK_TCK')
        except (ValueError, TypeError, AttributeError):
            pass

        # This is used to test if we can access /proc.
        self._btime = 0
        try:
            self._btime = self._boot_time()
        except IOError:
            pass

        major, minor, patchlevel = platform.python_version_tuple()
        info = {
            "version": platform.python_version(),
            "implementation": platform.python_implementation(),
            "major": major,
            "minor": minor,
            "patchlevel": patchlevel
        }

        self.process_metrics = {
            "info": Gauge(
                ns + "python_info",
                "Python platform information."
            ),
            "vmem": Gauge(
                ns + "process_virtual_memory_bytes",
                "Virtual memory size in bytes."
            ),
            "rss": Gauge(
                ns + "process_resident_memory_bytes",
                "Resident memory size in bytes."
            ),
            "start_time": Gauge(
                ns + "process_start_time_seconds",
                "Start time of the process since unix epoch in seconds."
            ),
            "cpu": Counter(
                ns + "process_cpu_seconds_total",
                "Total user and system CPU time spent in seconds."
            ),
            "open_fds": Gauge(
                ns + "process_open_fds",
                "Number of open file descriptors."
            ),
            "max_fds": Gauge(
                ns + "process_max_fds",
                "Maximum number of open file descriptors."
            )
        }

        # Only include these metrics if CPython and gc supports get_stats
        if hasattr(gc, 'get_stats') and platform.python_implementation() == 'CPython':
            self.process_metrics["collected"] = Counter(
                ns + "python_gc_objects_collected",
                "Objects collected during gc."
            )
            self.process_metrics["uncollectable"] = Counter(
                ns + "python_gc_objects_uncollectable",
                "Uncollectable object found during GC."
            )
            self.process_metrics["collections"] = Counter(
                ns + "python_gc_collections",
                "Number of times this generation was collected."
            )

        self.process_metrics["info"].set(info, 1.0)

    def _boot_time(self):
        with open("/proc/stat", 'rb') as stat:
            for line in stat:
                if line.startswith(b'btime '):
                    return float(line.split()[1])

    def values(self):
        return self.process_metrics.values()

    def collect(self):
        """
        Update the metrics from the latest system info in /proc.
        Although this method uses open and read it shouldn't pose any blocking
        issues for asyncio as we are only accessing the in-memory procfs
        https://en.wikipedia.org/wiki/Procfs
        """
        if not self._btime:
            return

        try:
            with open("/proc/self/stat", 'rb') as stat:
                parts = (stat.read().split(b')')[-1].split())

            self.process_metrics["vmem"].set("", float(parts[20]))
            self.process_metrics["rss"].set("", float(parts[21]) * _PAGESIZE)
            start_time_secs = float(parts[19]) / self._ticks
            self.process_metrics["start_time"].set(
                "", start_time_secs + self._btime
            )
            utime = float(parts[11]) / self._ticks
            stime = float(parts[12]) / self._ticks
            self.process_metrics["cpu"].set("", utime + stime)
        except IOError:
            pass

        try:
            with open("/proc/self/limits", 'rb') as limits:
                for line in limits:
                    if line.startswith(b'Max open file'):
                        self.process_metrics["max_fds"].set(
                            "", float(line.split()[3])
                        )
                        break

            self.process_metrics["open_fds"].set(
                "", float(len(os.listdir("/proc/self/fd")))
            )
        except (IOError, OSError):
            pass

        # Update gc metrics if enabled.
        if "collected" in self.process_metrics:
            for generation, stat in enumerate(gc.get_stats()):
                generation = {"generation": str(generation)}
                self.process_metrics["collected"].set(
                    generation, stat["collected"]
                )
                self.process_metrics["uncollectable"].set(
                    generation, stat["uncollectable"]
                )
                self.process_metrics["collections"].set(
                    generation, stat["collections"]
                )




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
Provides a JMS-like Connection/Session/Producer/Consumer/Message abstraction for
AMQP 0.9.1 connections (to RabbitMQ, though may work with other brokers).

By using JMS-like semantics the intention is to have an API that is already
reasonably abstracted from the underlying messaging fabric, so that it should
*hopefully* be a little simpler to build implementations for other fabrics.
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3

import json

# Tested using Pika 1.0.1, may not work correctly with earlier versions.
import pika  # sudo pip3 install pika

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.messaging_exceptions import *

class Connection(object):
    def __init__(self, url="amqp://localhost:5672"):
        """
        Creates a connection. A newly created connection must be opened with the
        Connection.open() method before it can be used.

        For RabbitMQ AMQP Connection URL documentation see:
        https://pika.readthedocs.io/en/stable/modules/parameters.html
        https://pika.readthedocs.io/en/stable/examples/using_urlparameters.html
        """
        self.logger = init_logging(log_name="amqp_0_9_1_messaging")
        self.logger.info("Creating Connection with url: {}".format(url))
        self.parameters  = pika.URLParameters(url)
        self.parameters.credentials = pika.PlainCredentials('guest','guest')


    def open(self, timeout=None):
        """
        Opens the connection.
        """
        if hasattr(self, "connection") and self.connection.is_open:
            raise ConnectionError(
                "Connection to {host}:{port} is already open".format(
                    host=self.parameters.host, port=self.parameters.port
                )
            )

        self.logger.info(
            "Opening Connection to {host}:{port}".format(
                host=self.parameters.host, port=self.parameters.port
            )
        )
        try:
            self.connection = pika.BlockingConnection(self.parameters)
            self.logger.info('Connection is Open:'+str(self.connection.is_open))

        except pika.exceptions.AMQPConnectionError as e:
            raise ConnectionError(repr(e))

    def is_open(self):
        """
        Return True if the connection is open, False otherwise. 
        """ 
        return hasattr(self, "connection") and self.connection.is_open

    def close(self):
        """
        Closes the connection.
        """
        if self.is_open():
            self.logger.info(
                "Closing Connection to {host}:{port}".format(
                    host=self.parameters.host, port=self.parameters.port
                )
            )
            self.connection.close()

    def session(self, name=None, transactional=False, auto_ack=False):
        """
        Creates a Session object.

        TODO add code so connection can store and retrieve sessions by name.
        TODO with JMS calling createSession should create a Session instance
        and each Session's Consumer MessageListener will run in a different
        thread. Pika is single threaded and moreover generally not thread-safe
        so to have similar behaviour it is (generally) necessary to have a
        separate Pika *connection* per thread. We should threfore create a new
        connection for each session, but for now only support a single Session
        such that the first call to session() will create the Session and
        subsequent calls will return that Session.
        """
        if not hasattr(self, "_session"):
            self._session = Session(self, name, transactional, auto_ack)
        return self._session

    def set_timeout(self, callback, delay):
        """
        Executes the specified callback function after the specified ms delay.
        Intended to have the same semantics as the JavaScript setTimeout().
        NOTE: the timer callbacks are dispatched only in the scope of 
        BlockingConnection.process_data_events() and
        BlockingChannel.start_consuming() e.g. after the start() method below
        has been called.
        Clamp delay to >= 0 as call_later doesn't handle negative values.
        """
        if delay < 0:
            delay = 0
        return self.connection.call_later(delay / 1000, callback)

    def clear_timeout(self, timeout_id):
        """
        Remove a timer if it’s still in the timeout stack.
        Intended to have the same semantics as the JavaScript clearTimeout().
        """
        self.connection.remove_timeout(timeout_id)

    def start(self):
        """
        Starts (or restarts) a connection's delivery of incoming messages.
        A call to start on a connection that has already been started is ignored.

        TODO when multiple sessions are supported start them all.
        There is actually a lot going on in start_consuming as Pika is basically
        single threaded and largely not thread safe the main docs are here
        https://pika.readthedocs.io/en/stable/modules/adapters/blocking.html

        One issue is how to communicate with the Pika dispatch loop from a
        different thread. I *think* that the way to do it is via the
        connection.add_callback_threadsafe(callback) method which requests a
        call to the given function as soon as possible in the context of this
        connection’s thread.

        It's also not clear how to consume on multiple channels as the
        start_consuming call is a method on channel, but it blocks, so how to
        consume on other channels in that case??
        """
        if hasattr(self, "_session"):
            try:
                self._session.channel.start_consuming()
            except pika.exceptions.ConnectionClosedByBroker as e:
                raise ConnectionError(repr(e))

# ------------------------------------------------------------------------------

class Session(object):
    def __init__(self, connection, name, transactional, auto_ack):
        """
        Sessions provide a context for sending and receiving Messages. Messages
        are sent and received using the Producer and Consumer objects associated
        with a Session.

        Each Producer and Consumer is created by supplying either a target or
        source address to the producer and consumer methods of the Session. 
        The address is supplied via a string syntax.
        """
        if not connection.is_open():
            raise SessionError("Unable to create Session as Connection is not open")

        connection.logger.info("Creating Session")  # TODO log name
        self.connection = connection
        self.name = 'task'  # TODO do something useful with name
        # TODO maybe do something with transactional?
        self.auto_ack = auto_ack

        """
        For RabbitMQ AMQP Channel documentation see:
        https://pika.readthedocs.io/en/stable/modules/channel.html
        """
        self.channel = connection.connection.channel()

    def acknowledge(self, message=None, threadsafe=False):
        """
        Acknowledge the given Message. If message is None, then all 
        unacknowledged messages on the session are acknowledged.

        https://www.rabbitmq.com/confirms.html

        param message: the message to acknowledge or None
        type message: Message

        basic_ack(delivery_tag=0, multiple=False)
        If multiple is set to True, the delivery tag is treated as “up to and
        including”, so that multiple messages can be acknowledged with a single
        method. If set to False, the delivery tag refers to a single message.

        If the multiple field is True, and the delivery tag is zero, this 
        indicates acknowledgement of all outstanding messages.
        """
        if message == None:
            def ack():
                """
                The main purpose of this nested function is to make it fairly easy
                to either directly call the underlying pika basic_ack or defer
                to connection.add_callback_threadsafe as a callback.
                """
                self.channel.basic_ack(delivery_tag=0, multiple=True)

            if threadsafe:
                # If called from a foreign thread add ack to the event queue
                # as this is the only thread safe Pika API call.
                self.channel.connection.add_callback_threadsafe(ack)
            else:
                ack()
        else:
            # Only acknowledge the specific message.
            message.acknowledge(multiple=False, threadsafe=threadsafe)

    def recover(self, requeue=False):
        """
        Restarts message delivery with the oldest unacknowledged message. 
        This method asks the server to redeliver all unacknowledged messages on
        a specified channel. Zero or more messages may be redelivered.
        """
        self.channel.basic_recover(requeue)

    def is_open(self):
        """
        Return True if the session is open, False otherwise. 
        """ 
        return self.channel.is_open

    def close(self):
        """
        Closes the session.
        """
        if self.is_open():
            self.logger.info("Closing Session")  # TODO log session name
            self.channel.close()

    def producer(self, target=""):
        """
        Creates a Producer used to send messages to the specified target.
        :param target: The target to which messages will be sent
        :type target: str
        """
        return Producer(self, target)

    def consumer(self, source=""):
        """
        Creates a Consumer used to fetch messages from the specified source.
        :param source: The source of messages
        :type source: str
        """
        return Consumer(self, source)

# ------------------------------------------------------------------------------

class Destination(object):
    def __init__(self):
        """
        For RabbitMQ AMQP Channel documentation see:
        https://pika.readthedocs.io/en/stable/modules/channel.html
        Default values taken from exchange_declare, queue_declare, queue_bind
        """
        self.declare = {
            "queue": "",
            "exchange": "",
            "exchange-type": "direct",
            "passive": False,
            "internal": False,
            "durable": False,
            "exclusive": False,
            "auto-delete": False,
            "arguments": None,
        }

        # Defaults for subscription queues (exclusive and autodelete True)
        self.link_declare = {
            "queue": "",
            "passive": False,
            "internal": False,
            "durable": False,
            "exclusive": True,
            "auto-delete": True,
            "arguments": None,
        }

        """
        The x-subscribe map of a link controls the exclusive and arguments
        fields of a subscription, which relates to the AMQP 0.9.1 basic.consume
        or the AMQP 0.10 message.subscribe protocol commands. This is mainly
        used to request an exclusive subscription. This prevents other
        subscribers from subscribing to the queue.
        """
        self.link_subscribe = {
            "exclusive": False,
            "arguments": None,
        }

        self.bindings = []

    def parse_address(self, address):
        """
        Parses an address string with the following format:

        <name> [ / <subject> ] [ ; <options> ]

        Where options is of the form: { <key> : <value>, ... }

        And values may be numbers, strings, maps (dictionaries) or lists

        The options map permits the following parameters:

        <name> [ / <subject> ] ; {
            node: {
                type: queue | topic,
                durable: True | False,
                auto-delete: True | False,
                x-declare: { ... <declare-overrides> ... },
                x-bindings: [<binding_1>, ... <binding_n>]
            },
            link: {
                name: <link-name>,
                durable: True | False,
                reliability: unreliable | at-most-once | at-least-once | exactly-once,
                x-declare: { ... <declare-overrides> ... },
                x-bindings: [<binding_1>, ... <binding_n>]
            }
        }

        The node refers to the AMQP node e.g. a queue or exchange being referred
        to in the address, whereas the link allows configuration of a logical
        "subscriber queue" that will be created when the address node is an
        exchange such as, for example, news-service/sports.

        Bindings are specified as a map with the following options:

        {
            exchange: <exchange>,
            queue: <queue>,
            key: <key>,
            arguments: <arguments>
        }

        The x-declare map permits protocol specific keys and values to be
        specified when exchanges or queues are declared. These keys and
        values are passed through when creating a node or asserting facts
        about an existing node.

        For Producers the node implicitly defaults to a topic and for Consumers
        it implicitly defaults to a queue, but this may be overridden by setting
        exchange or queue in the x-declare object or by setting node type to
        queue or topic.

        Examples:
        myqueue; {"node": {"x-declare": {"durable": true, "exclusive": true, "auto-delete": true}}}'
        myqueue; {"node": {"x-declare": {"exchange": "test-headers", "exchange-type": "headers", "durable": true, "auto-delete": true}}}

        myqueue; {"node": {"x-declare": {"durable": true, "auto-delete": true}, "x-bindings": [{"exchange": "amq.match", "queue": "myqueue", "key": "data1", "arguments": {"x-match": "all", "data-service": "amqp-delivery", "item-owner": "Sauron"}}]}}

        myqueue; {"node": {"durable": true, "x-bindings": [{"exchange": "amq.match", "queue": "myqueue", "key": "data1", "arguments": {"x-match": "all", "data-service": "amqp-delivery", "item-owner": "Sauron"}}, {"exchange": "amq.match", "queue": "myqueue", "key": "data2", "arguments": {"x-match": "all", "data-service": "amqp-delivery", "item-owner": "Gandalf"}}]}}

        myqueue; {"node": {"x-declare": {"durable": true, "auto-delete": false}}, "link": {"x-subscribe": {"exclusive": true}}}'

        news-service/sports

        news-service/sports; {"node": {"x-declare": {"exchange": "news-service", "exchange-type": "topic"}}}

        news-service/sports; {"node": {"x-declare": {"exchange": "news-service", "exchange-type": "topic", "auto-delete": true}}, "link": {"x-declare": {"queue": "news-queue", "exclusive": false}}}

        Topic exchanges can be declared by message producers as follows:
        ; {"node": {"x-declare": {"exchange": "news-service", "exchange-type": "topic"}}}

        For the case where the address comprises just the options string
        the leading semicolon is optional e.g. the following is also valid:
        {"node": {"x-declare": {"exchange": "news-service", "exchange-type": "topic"}}}

        """
        kv = address.split(";")
        options_string = kv[1] if len(kv) == 2 else "{}"
        kv = kv[0].split("/")
        self.subject = kv[1].strip() if len(kv) == 2 else ""
        self.name = kv[0].strip()
        
        # Handle case where address comprises just the options string.
        if len(self.name) >= 2 and self.name[0] == "{":
            options_string = self.name
            self.name = ""

        options = json.loads(options_string)

        # print(options_string)
        # print(options)
        # print(self.subject)
        # print(self.name)

        node = options.get("node")
        if node:
            x_declare = node.get("x-declare")
            if x_declare and type(x_declare) == type(self.declare):
                self.declare.update(x_declare)
                if self.name:
                    # If node type is set then set queue or exchange name in
                    # declare if not already explicitly set in x-declare
                    if node.get("type") == "queue" and not self.declare.get("queue"):
                        self.declare["queue"] = self.name
                    if node.get("type") == "topic" and not self.declare.get("exchange"):
                        self.declare["exchange"] = self.name
                else:
                    # Handle case where address comprises just the options string.
                    if node.get("type") == "queue":
                        self.name = self.declare.get("queue", "")
                    if node.get("type") == "topic":
                        self.name = self.declare.get("exchange", "")
                    # Handle edge cases where node type is not explicitly set
                    if not self.name:
                        self.name = self.declare.get("exchange", "")
                    if not self.name:
                        self.name = self.declare.get("queue", "")

            # Can set durable and auto-delete on node as a shortcut if we don't
            # need any other declare overrides.
            if node.get("durable"):
                self.declare["durable"] = True
            if node.get("auto-delete"):
                self.declare["auto-delete"] = True
            # If x-bindings set populate Destination's bindings
            x_bindings = node.get("x-bindings")
            if x_bindings and type(x_bindings) == type(self.bindings):
                self.bindings = x_bindings

        link = options.get("link")
        if link:
            x_declare = link.get("x-declare")
            if x_declare and type(x_declare) == type(self.link_declare):
                self.link_declare.update(x_declare)
            x_subscribe = link.get("x-subscribe")
            if x_subscribe and type(x_subscribe) == type(self.link_subscribe):
                self.link_subscribe.update(x_subscribe)

# ------------------------------------------------------------------------------

class Producer(Destination):
    def __init__(self, session, target):
        """
        A client uses a Producer object to send messages to a destination.
        """
        super().__init__()  # Call Destination constructor

        session.connection.logger.info(
            "Creating Producer with address: {}".format(target)
        )
        self.session = session

        """
        Confirmation delivery_tags/sequence numbers start at 1
        https://www.rabbitmq.com/confirms.html#publisher-confirms
        """
        self.next_publish_seq_no = 1
        self.saved_ack_seq_no = 0
        self.saved_nack_seq_no = 0
        self.saved_seq_no = 0
        self.previous_nack_start = 1

        """
        self.undelivered is used when enable_exceptions(sync=False) has been
        called. It records the sequence numbers of undelivered (unpublished)
        messages as tuples, representing contiguous blocks of unpublished
        messages. This information will be recorded in an exception thrown by
        send() and may be used by applications to identify messages to resend.
        """
        self.undelivered = []

        try:
            self.parse_address(target)
        except ValueError as e:
            raise ProducerError("Failed to parse address: {} {}".format(target, e))

        # Check if an exchange with the name of this destination exists.
        if self.name:
            try:
                # Use temporary channel as the channel gets closed on an exception.
                temp_channel = self.session.connection.connection.channel()
                temp_channel.exchange_declare(self.name, passive=True)
                temp_channel.close()
            except pika.exceptions.ChannelClosedByBroker as e:
                session.connection.logger.info(e)
                # If 404 NOT_FOUND the specified exchange doesn't exist.
                if e.reply_code == 404:
                    # If no exchange declared assume default direct exchange.
                    if self.name != self.declare.get("exchange"):
                        self.subject = self.name
                        self.name = ""  # Set exchange to default direct.

        if self.declare.get("exchange"):
            # Exchange declare.
            self.session.channel.exchange_declare(
                exchange=self.declare["exchange"],
                exchange_type=self.declare["exchange-type"],
                passive=self.declare["passive"],
                durable=self.declare["durable"],
                auto_delete=self.declare["auto-delete"],
                arguments=self.declare["arguments"],
            )
        # print("name = " + self.name)
        # print("subject = " + self.subject)

    def enable_exceptions(self, sync=False):
        """
        With JMS the spec tends towards implying a delivery guarantee, in
        particular if messages are marked persistent. However, most AMQP
        implementations publish asynchronously and have buffering to reduce
        latencies due to protocol round-trips. To meet JMS delivery guarantees
        implementations might allow a sync_publish option, which would wait
        for the basic_ack from the broker. Otherwise any exceptions on send,
        if any, might occur *after* the message that actually caused a failure.

        By default the send() method below publishes asynchronously and does
        *not* raise exceptions unless enabled by calling enable_exceptions().

        When enable_exceptions() is called the default behaviour is to raise an
        exception on the first call to send() after one or more basic_nack has
        been received from the broker.

        The exception contains the sequence numbers of the nacked messages and
        this information may be used by applications to identify messages that
        might need to be resent.

        If sync=True send() will instead block until delivery confirmation by
        the broker. This mode of operation is what the BlockingChannel
        confirm_delivery() does and the underlying basic_publish call will block
        until ack or nack is received. NOTE that setting sync=True could impact
        message throughput and latency considerably and unfortunately the pika
        BlockingChannel doesn't provide a convenient way to block for a batch
        of messages.

        WARNING: do not modify the exception_ack_nack_callback function unless
        you know what you are doing. The gist is that it records the sequence
        numbers of the ack and nack methods returned by the broker.

        The possibility of nacks filling "holes" in the sequencing is part of
        the reason the exception callback is quite complicated. The function
        is basically creating a list of tuples holding each range of undelivered
        messages and we process the Ack method to find self.previous_nack_start
        That is used if the nack sequence number arrives out of sequence so
        we can identify the "hole" to fill.

        One reason to be careful is that the out of sequence case is sporadic
        and only cropped up on edge cases.
        """

        def exception_ack_nack_callback(result):
            method = result.method
            seq_no = method.delivery_tag

            if isinstance(method, pika.spec.Basic.Ack):
                if not method.multiple:
                    prev_seq_no = seq_no
                elif self.saved_nack_seq_no > self.saved_ack_seq_no:
                    prev_seq_no = self.saved_nack_seq_no
                else:
                    prev_seq_no = self.saved_ack_seq_no
                                                        
                if self.saved_seq_no != prev_seq_no - 1:
                    self.previous_nack_start = self.saved_seq_no + 1
                self.saved_seq_no = seq_no

                self.saved_ack_seq_no = seq_no + 1
            elif isinstance(method, pika.spec.Basic.Nack):
                if seq_no < self.saved_ack_seq_no:
                    prev_seq_no = self.previous_nack_start
                else:
                    prev_seq_no = self.saved_ack_seq_no

                if self.undelivered:
                    last = self.undelivered[-1]
                    if last[0] == prev_seq_no:
                        self.undelivered[-1] = (prev_seq_no, seq_no)
                    else:
                        self.undelivered.append((prev_seq_no, seq_no))
                else:
                    self.undelivered.append((prev_seq_no, seq_no))

                self.saved_nack_seq_no = seq_no + 1

        if sync:
            self.session.channel.confirm_delivery()
        else:
            self.saved_ack_seq_no = 1
            self.saved_nack_seq_no = 1

            # Need to access the blocking channel's underlying async channel
            # for this mode of operation.
            self.session.channel._impl.confirm_delivery(
                ack_nack_callback=exception_ack_nack_callback,
            ) 

    def send(self, message, threadsafe=False):
        """
        For RabbitMQ AMQP Channel documentation see:
        https://pika.readthedocs.io/en/stable/modules/channel.html
        https://pika.readthedocs.io/en/stable/modules/spec.html#pika.spec.BasicProperties

        basic_publish(exchange, routing_key, body, properties=None,
                      mandatory=False)
        Delivery mode 2 makes the broker save the message to disk.
        """

        def publish():
            """
            The main purpose of this nested function is to make it fairly easy
            to either directly call the underlying pika basic_publish or defer
            to connection.add_callback_threadsafe as a callback.
            """

            # If message.subject is set use that as the routing_key, otherwise use
            # the Producer target default subject parsed from address string.
            subject = message.subject
            routing_key = subject if subject else self.subject

            # If message.expiration is set to an invalid value (like a
            # non-numeric or a negative value) we "clamp" it to a "0"
            clamped_expiration = None
            if message.expiration is not None:
                try:
                    clamped_expiration=str(int(float(message.expiration)))
                    if clamped_expiration.startswith("-"):
                        clamped_expiration = "0"
                except ValueError as e:
                    clamped_expiration = "0"

            properties = pika.BasicProperties(
                headers=message.properties,
                content_type=message.content_type,
                content_encoding=message.content_encoding,
                delivery_mode=2 if message.durable else 1,
                priority=message.priority,
                correlation_id=message.correlation_id,
                reply_to=message.reply_to,
                expiration=clamped_expiration,
                message_id=message.message_id,
                timestamp=message.timestamp,
                type=message.type,
                user_id=message.user_id,
                app_id=message.app_id,
                cluster_id=message.cluster_id,
            )

            try:
                self.session.channel.basic_publish(
                    exchange=self.name,
                    routing_key=routing_key,
                    body=message.body,
                    properties=properties,
                )
            except pika.exceptions.NackError as e:
                undelivered = [(self.next_publish_seq_no, self.next_publish_seq_no)]
                send_exception = SendError(
                    "Failed to send message: undelivered = {}".format(undelivered)
                )
                send_exception.undelivered = undelivered
                raise send_exception

        if self.undelivered:  # Will only be True if enable_exceptions(sync=False).
            """
            If enable_exceptions() has been called with sync=False (the default)
            the sequence numbers of nacked messages are recorded in undelivered,
            which is a list of tuples each of which records the first and last
            sequence number of groups of messages that failed to be published.
            There may be multiple items if groups of nacked messages are non-
            contiguous. If there are any undelivered/unpublished messages an
            exception is thrown which contains a copy of this list of tuples.
            """
            send_exception = SendError(
                "Failed to send messages: undelivered = {}".format(
                    self.undelivered
                )
            )
            send_exception.undelivered = self.undelivered.copy()
            self.undelivered = []
            raise send_exception
        elif threadsafe:
            self.session.connection.connection.add_callback_threadsafe(publish)
        else:
            publish()

        self.next_publish_seq_no += 1

# ------------------------------------------------------------------------------

class Consumer(Destination):
    def __init__(self, session, source):
        """
        A client uses a Consumer object to receive messages from a destination.
        """
        super().__init__()  # Call Destination constructor

        session.connection.logger.info(
            "Creating Consumer with address: {}".format(source)
        )
        self.session = session

        # Set default capacity/message prefetch to 500
        self._capacity = 500
        self.session.channel.basic_qos(prefetch_count=self._capacity)

        try:
            self.parse_address(source)
        except ValueError as e:
            raise ConsumerError("Failed to parse address: {} {}".format(source, e))

        # Check if an exchange with the name of this destination exists.
        exchange = None
        if self.name:
            try:
                # Use temporary channel as the channel gets closed on an exception.
                temp_channel = self.session.connection.connection.channel()
                temp_channel.exchange_declare(self.name, passive=True)
                temp_channel.close()
                exchange = self.name
            except pika.exceptions.ChannelClosedByBroker as e:
                if self.subject:
                    if self.name == self.declare.get("exchange"):
                        exchange = self.name
                    else:
                        raise ConsumerError(e.reply_text)
                # Otherwise we assume default direct exchange

        if exchange:  # Is this address an exchange?
            # Destination is an exchange, create subscription queue and
            # add binding between exchange and queue with subject as key.
            if self.declare.get("queue"):
                self.name = self.declare.get("queue")
            elif self.link_declare.get("queue"):
                self.name = self.link_declare.get("queue")
            else:
                self.name = ""

            if len(self.bindings) == 0 and self.subject:
                self.bindings.append(
                    {"queue": self.name, "exchange": exchange, "key": self.subject}
                )

        # Declare queue, exchange and bindings as necessary
        if self.declare.get("exchange"):
            # Exchange declare - unusual scenario for Consumer, but handle it.
            self.session.channel.exchange_declare(
                exchange=self.declare["exchange"],
                exchange_type=self.declare["exchange-type"],
                passive=self.declare["passive"],
                durable=self.declare["durable"],
                auto_delete=self.declare["auto-delete"],
                arguments=self.declare["arguments"],
            )
        # Queue declare
        declare = self.declare
        if exchange and not self.declare.get("queue"):
            declare = self.link_declare
        if self.name == "":
            declare["auto-delete"] = True

        result = self.session.channel.queue_declare(
            queue=self.name,
            passive=declare["passive"],
            durable=declare["durable"],
            exclusive=declare["exclusive"],
            auto_delete=declare["auto-delete"],
            arguments=declare["arguments"],
        )
        """
        Get the queue name from the result of the queue_declare to deal with
        the case of server created names when we pass queue="" to queue_declare
        see https://www.rabbitmq.com/tutorials/tutorial-three-python.html
        """
        self.name = result.method.queue

        for binding in self.bindings:
            if binding["exchange"] == "":
                continue  # Can't bind to default
            self.session.channel.queue_bind(
                queue=binding["queue"],
                exchange=binding["exchange"], 
                routing_key=binding.get("key"),
                arguments=binding.get("arguments"),
            )

    def set_message_listener(self, message_listener):
        """
        For RabbitMQ AMQP Channel documentation see:
        https://pika.readthedocs.io/en/stable/modules/channel.html

        basic_consume(queue, on_message_callback, auto_ack=False,
                      exclusive=False, consumer_tag=None, arguments=None,
                      callback=None)

        To specify an exclusive subscription to a queue use an address like:
        myqueue; {"node": {"x-declare": {"durable": true, "auto-delete": false}}, "link": {"x-subscribe": {"exclusive": true}}}'
        """
        ex = self.link_subscribe.get("exclusive", False)
        args = self.link_subscribe.get("arguments", None)
        self._message_listener = message_listener
        try:
            self.session.channel.basic_consume(
                on_message_callback=self.message_listener,
                queue=self.name,
                auto_ack=self.session.auto_ack,
                exclusive=ex,
                arguments=args
            )
        except pika.exceptions.ChannelClosedByBroker as e:
            raise ConsumerError(e.reply_text)

    def message_listener(self, channel, method, properties, body):
        """
        This is the Consumer's default message listener, its job is to create
        a Message instance that encapsulates some of the AMQP details into
        something more akin to a JMS Message, then delegate to the registered
        message listener.

        https://pika.readthedocs.io/en/stable/modules/spec.html#pika.spec.Basic.Deliver
        https://pika.readthedocs.io/en/stable/modules/spec.html#pika.spec.BasicProperties

        channel: pika.Channel
        method: pika.spec.Basic.Deliver
        properties: pika.spec.BasicProperties
        body: bytes
        """
        if hasattr(self, "_message_listener"):
            # Now call the registered message listener
            message = Message(
                body,
                properties=properties.headers,
                content_type=properties.content_type,
                content_encoding=properties.content_encoding,
                redelivered=method.redelivered,
                durable=(properties.delivery_mode == 2),
                priority=properties.priority,
                correlation_id=properties.correlation_id,
                reply_to=properties.reply_to,
                expiration=properties.expiration,
                message_id=properties.message_id,
                timestamp=properties.timestamp,
                type=properties.type,
                user_id=properties.user_id,
                app_id=properties.app_id,
                cluster_id=properties.cluster_id,
            )
            # These two private attributes are added to Message to enable
            # Message's acknowledge() methods
            message._channel = channel
            message._delivery_tag = method.delivery_tag
            self._message_listener(message)

    @property
    def capacity(self):
        return self._capacity

    @capacity.setter
    def capacity(self, capacity):
        self._capacity = capacity
        self.session.channel.basic_qos(prefetch_count=capacity)

# ------------------------------------------------------------------------------

class Message(object):
    def __init__(
        self,
        body="",  # Pika expects empty string not None for no body.
        properties=None,
        content_type=None,
        content_encoding=None,
        redelivered=False,
        durable=True,
        priority=None,
        correlation_id=None,
        reply_to=None,
        expiration=None,
        message_id=None,
        timestamp=None,
        type=None,
        user_id=None,
        app_id=None,
        cluster_id=None,
        subject=None,
    ):
        """
        Provides an abstraction for messages comprising a body, application
        properties and a set of headers used by the messaging fabric to identify
        and route messages and provide additional metadata.
        """
        self.body = body
        self.properties = properties  # Holds application property key/value pairs

        # Values below from:
        # pika.BasicProperties(delivery_mode=2)
        # https://pika.readthedocs.io/en/stable/modules/spec.html#pika.spec.BasicProperties
        # https://stackoverflow.com/questions/18403623/rabbitmq-amqp-basicproperties-builder-values
        self.content_type = content_type
        self.content_encoding = content_encoding
        self.redelivered = redelivered
        self.durable = durable
        self.priority = priority
        self.correlation_id = correlation_id
        self.reply_to = reply_to
        self.expiration = expiration
        self.message_id = message_id
        self.timestamp = timestamp
        self.type = type
        self.user_id = user_id
        self.app_id = app_id
        self.cluster_id = cluster_id

        if self.properties == None:
            self.properties = {}
        self.subject = subject  # Set subject *after* self.properties initialised.

    def __repr__(self):
        args = []
        for name in [
            "content_type",
            "content_encoding",
            "priority",
            "message_id",
            "type",
            "user_id",
            "app_id",
            "cluster_id",
            "reply_to",
            "correlation_id",
            "expiration",
            "timestamp",
            "redelivered",
            "durable",
            "properties",
        ]:
            value = self.__dict__[name]
            if value is not None:
                args.append("%s=%r" % (name, value))
        if self.subject:
            args.append("subject=%r" % self.subject)
        if self.body is not None:
            if args:
                args.append("body=%r" % self.body)
            else:
                args.append(repr(self.body))
        return "Message(%s)" % ", ".join(args)

    """
    Unlike AMQP 1.0 AMQP 0.9.1 doesn't have a Message subject, but it is a
    useful concept as often we wish to publish messages to a generic producer
    Node such as a topic exchange and have messages delivered based on their
    subject. Although basic_publish allows one to achieve the same result it is
    much more coupled with AMQP 0.9.1 (and AMQP 0.10) protocol details than
    if it were a property of the Message, which is also more intuitive.
    We map "subject" to a message property that is unlikely to collide with any
    application set message properties, this will be used by the Producer.send()
    but will also be available to consumers as a Message property.
    """

    @property
    def subject(self):
        return self.properties.get("x-amqp-0-9-1.subject")

    @subject.setter
    def subject(self, subject):
        if subject:
            self.properties["x-amqp-0-9-1.subject"] = subject
            # print(self.properties)

    def acknowledge(self, multiple=True, threadsafe=False):
        """
        Acknowledge this Message.

        Messages that have been received but not acknowledged may be redelivered.

        basic_ack(delivery_tag=0, multiple=False)
        If multiple is set to True, the delivery tag is treated as “up to and
        including”, so that multiple messages can be acknowledged with a single
        method. If set to False, the delivery tag refers to a single message.

        If the multiple field is True, and the delivery tag is zero, this 
        indicates acknowledgement of all outstanding messages.
        """
        if hasattr(self, "_channel"):
            def ack():
                """
                The main purpose of this nested function is to make it fairly easy
                to either directly call the underlying pika basic_ack or defer
                to connection.add_callback_threadsafe as a callback.
                """
                if multiple:
                    """
                    The multiple == True branch behaves like JMS Message.acknowledge()
                    https://docs.oracle.com/javaee/7/api/javax/jms/Message.html#acknowledge--
                    By invoking acknowledge on a consumed message, a client
                    acknowledges all messages consumed by the session that the
                    message was delivered to. 
                    """
                    self._channel.basic_ack(delivery_tag=0, multiple=True)
                else:
                    self._channel.basic_ack(delivery_tag=self._delivery_tag)

            if threadsafe:
                # If called from a foreign thread add ack to the event queue
                # as this is the only thread safe Pika API call.
                self._channel.connection.add_callback_threadsafe(ack)
            else:
                ack()




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 workers_blocking_asyncio.py
# PYTHONPATH=.. LOG_LEVEL=DEBUG python3 workers_blocking_asyncio.py
#
"""
This is an asyncio based version of the workers_blocking.py example.

The main loop uses asyncio, and a "handler_task" representing a long running
job that would block the event loop if called directly is launched as a Task
(Executor thread) by the main message_listener via a loop.run_in_executor call.
"""

import sys
assert sys.version_info >= (3, 6) # Bomb out if not running Python3.6

import asyncio, json, time, opentracing

import atexit, concurrent.futures, contextlib

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer, span_context, inject_span
from asl_workflow_engine.amqp_0_9_1_messaging_asyncio import Connection, Message
from asl_workflow_engine.messaging_exceptions import *

class Worker():

    def __init__(self, name):
        self.name = name
        # Initialise logger
        self.logger = init_logging(log_name=name)

        # Note that although we will use asyncio await loop.run_in_executor
        # to actually launch the tasks, the actual ThreadPoolExecutor itself
        # is not asyncio so we use ExitStack, enter_context and close rather
        # than AsyncExitStack, enter_async_context and aclose as we would for
        # an asyncio context manager.
        context_stack = contextlib.ExitStack()
        # Create the ThreadPoolExecutor here because creating it in the
        # message listener/handler is really expensive. Also running it as a
        # context manager using a with statement there won't actually sort the
        # issue as the context manager blocks until the threads exit so will
        # block the event loop in the main thread which is exactly what we
        # *don't* want to do.
        self.executor = context_stack.enter_context(
            concurrent.futures.ThreadPoolExecutor(max_workers=10)
        )

        # Note: exit_handler is a nested closure which "decouples" the lifetime
        # of the exit_handler from the lifetimes of instances of the class.
        # https://stackoverflow.com/questions/16333054/what-are-the-implications-of-registering-an-instance-method-with-atexit-in-pytho
        @atexit.register
        def exit_handler():
            print("exit_handler called")
            context_stack.close()
            print("exit_handler exiting")


    # Launched as an Executor Task from the message listener
    def handler_task(self, message):
        print(self.name + " handler_task started")
        print(message)

        with opentracing.tracer.start_active_span(
            operation_name=self.name,
            child_of=span_context("text_map", message.properties, self.logger),
            tags={
                "component": "workers",
                "message_bus.destination": self.name,
                "span.kind": "consumer",
                "peer.address": "amqp://localhost:5672"
            }
        ) as scope:
            # Delay for a while to simulate work then create simple reply.
            # In a real processor **** DO WORK HERE ****
            time.sleep(30)
            reply = {"reply": self.name + " reply"}

            """
            Create the response message by reusing the request note that this
            approach retains the correlation_id, which is necessary. If a fresh
            Message instance is created we would need to get the correlation_id
            from the request Message and use that value in the response message.
            """

            """
            Start an OpenTracing trace for the rpcmessage response.
            https://opentracing.io/guides/python/tracers/ standard tags are from
            https://opentracing.io/specification/conventions/
            """
            with opentracing.tracer.start_active_span(
                operation_name=self.name,
                child_of=opentracing.tracer.active_span,
                tags={
                    "component": "workers",
                    "message_bus.destination": message.reply_to,
                    "span.kind": "producer",
                    "peer.address": "amqp://localhost:5672"
                }
            ) as scope:
                message.properties=inject_span("text_map", scope.span, self.logger)
                message.subject = message.reply_to
                message.reply_to = None
                message.body = json.dumps(reply)
                #self.producer.send(message)
                #message.acknowledge() # Acknowledges the original request
                # These need to be thread safe if called from a thread
                # other than the main event loop thread.
                self.producer.send(message, threadsafe=True)
                message.acknowledge(threadsafe=True) # Acknowledges the original request

        print(self.name + " handler_task finished")

    async def message_listener(self, message):
        # Because we are basically launching the handler_task and doing nothing
        # with the result here (because the handler_task is acknowledging and
        # sending the result in the same way as the non-asyncio version) we
        # could actually make this method a subroutine (the underlying listener
        # autodetects if real listener is a subroutine or coroutine) and use the
        # self.executor.submit(self.handler_task, message)
        # below rather than loop.run_in_executor, but the approach here is
        # probably more idiomatic for code that is primarily asyncio.

        print(self.name + " message_listener called")
        loop = asyncio.get_event_loop()

        # N.B. whether to directly call the actual processing or use an
        # Executor *should be configurable*. For IO bound processing it may
        # make matters worse to launch in an Executor unless the processor
        # is itself blocking on other IO. In general asyncio is a better way
        # to enable high IO concurrency and Executors may be better for a
        # small number of more CPU bound processors.

        # Directly call the real delegated handler
        #self.handler_task(message)
        # Indirectly call the real delegated handler as an Executor Task
        #self.executor.submit(self.handler_task, message)
        await loop.run_in_executor(self.executor, self.handler_task, message)

    async def run(self):
        # Connect to worker queue and process data.
        connection = Connection("amqp://localhost:5672?connection_attempts=20&retry_delay=10&heartbeat=10")
        try:
            await connection.open()
            session = await connection.session()
            self.consumer = await session.consumer(
                self.name + '; {"node": {"auto-delete": true}}'
            )
            self.consumer.capacity = 100; # Enable consumer prefetch
            await self.consumer.set_message_listener(self.message_listener)
            self.producer = await session.producer()

            await connection.start(); # Wait until connection closes.
    
        except MessagingError as e:  # ConnectionError, SessionError etc.
            self.logger.error(e)

        connection.close();

if __name__ == '__main__':
    """
    Initialising OpenTracing here rather than in the Worker constructor as
    opentracing.tracer is a per process object not per thread.
    """
    create_tracer("workers", {"implementation": "Jaeger"})

    workers = ["SimpleBlockingProcessor"]
    loop = asyncio.get_event_loop()
    # Create list of Worker tasks using list comprehension
    tasks = [loop.create_task(Worker(name=w).run()) for w in workers]
    # Run the tasks concurrently
    loop.run_until_complete(asyncio.gather(*tasks))
    loop.close()
    



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 map3.py
#
"""
This test uses the AWS python SDK boto3 to access our local ASL Workflow Engine
https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/stepfunctions.html
https://docs.aws.amazon.com/code-samples/latest/catalog/code-catalog-python-example_code-stepfunctions.html

This example illustrates how to implement the behaviour of the iterate1.py
example using the ASL Map state.
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import boto3, time
from botocore.exceptions import ClientError

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer

ASL = """{
  "Comment": "State Machine to illustrate processing an array using an ASL Map state and passing each array item to a child Step Function. In this example InputPath, ItemsPath, MaxConcurrency and ResultPath are all omitted as we are using their defaults. The OutputPath is null, that means the input and Result are discarded, and the effective output from the state is an empty JSON object, {}",
  "StartAt": "ProcessArray",
  "States": {
    "ProcessArray": {
      "Type": "Map",
      "OutputPath": null,
      "Iterator": {
        "StartAt": "StepFunctionLauncher",
        "States": {
          "StepFunctionLauncher": {
            "Type": "Task",
            "Resource": "arn:aws:states:local:0123456789:states:startExecution",
            "Parameters": {  
                "Input.$": "$",
                "StateMachineArn": "arn:aws:states:local:0123456789:stateMachine:child_state_machine"
            },
            "Comment": "Launch the child Step Function, the input of which should be the *effective* input of the iteration.",
            "End": true
          }
        }
      },
      "End": true
    }
  }
}"""

child_ASL = """{
    "Comment": "Trivial Child Step Function",
    "StartAt": "StartState",
    "States": {
        "StartState": {
            "Type": "Pass",
            "End": true
        }
    }
}"""


items = ['[{"category": "reference", "author": "Nigel Rees", "title": "Sayings of the Century", "price": 8.95}, {"category": "fiction", "author": "Evelyn Waugh", "title": "Sword of Honour", "price": 12.99}, {"category": "fiction", "author": "Herman Melville", "title": "Moby Dick", "isbn": "0-553-21311-3", "price": 8.99}, {"category": "fiction", "author": "J. R. R. Tolkien", "title": "The Lord of the Rings", "isbn": "0-395-19395-8", "price": 22.99}]']

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name="map3")

    # Initialising OpenTracing. It's important to do this before the boto3.client
    # call as create_tracer "patches" boto3 to add the OpenTracing hooks.
    create_tracer("map3", {"implementation": "Jaeger"})

    # Initialise the boto3 client setting the endpoint_url to our local
    # ASL Workflow Engine
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session.client
    sfn = boto3.client("stepfunctions", endpoint_url="http://localhost:4584")
    state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:map3"
    child_state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:child_state_machine"
    
    def create_state_machines():
        # Create state machine using a dummy roleArn. If it already exists an
        # exception will be thrown, we handle that but ignore other exceptions.
        try:
            response = sfn.create_state_machine(
                name="map3", definition=ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )

            response = sfn.create_state_machine(
                name="child_state_machine", definition=child_ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )
        except sfn.exceptions.StateMachineAlreadyExists as e:
            #print(e.response)

            response = sfn.update_state_machine(
                stateMachineArn=state_machine_arn,
                definition=ASL
            )

            response = sfn.update_state_machine(
                stateMachineArn=child_state_machine_arn,
                definition=child_ASL
            )
        except ClientError as e:
            logger.error(e)


    create_state_machines()

    # Loop through items invoking a new state machine execution for each item
    for item in items:
        try:
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except sfn.exceptions.StateMachineDoesNotExist as e:
            logger.info(e)

            create_state_machines()
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except ClientError as e:
            logger.error(e)

    time.sleep(1)  # Give OpenTracing a chance to flush buffer before exiting




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
AWS resources are identified by Amazon Resource Names (ARNs) specified here:
http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html
This package comprises simple utility functions for creating and parsing ARNs.
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3


def create_arn(
    resource="",
    arn="arn",
    partition="aws",
    service="",
    region="",
    account="",
    resource_type=None,
):
    """
    Create an ARN string from its component parts, if this function is called
    passing a dictionary we use the ** operator to perform keyword expansion.
    """
    if isinstance(resource, dict):
        return create_arn(**resource)
    if resource_type:
        resource = resource_type + ":" + resource
    return "{}:{}:{}:{}:{}:{}".format(
        arn, partition, service, region, account, resource
    )

def parse_arn(arn):
    """
    Parse an ARN into a dictionary comprising the component parts of the ARN
    """
    elements = arn.split(":", 5)
    result = {
        "arn": elements[0],
        "partition": elements[1],
        "service": elements[2],
        "region": elements[3],
        "account": elements[4],
        "resource": elements[5],
        "resource_type": None,
    }
    if "/" in result["resource"]:
        result["resource_type"], result["resource"] = result["resource"].split("/", 1)
    elif ':' in result['resource']:
        result["resource_type"], result["resource"] = result["resource"].split(":", 1)
    return result
